@@ -730,6 +730,13 @@ class TestHttpd(_TestBase):
         assert b'400 Bad Request' in result, result
         assert b'500' not in result, result
 
+        sock = eventlet.connect(self.server_addr)
+        sock.sendall(b'GET / HTTP/1.0\r\nHost: localhost\r\nContent-length: -10\r\n\r\n')
+        result = recvall(sock)
+        assert result.startswith(b'HTTP'), result
+        assert b'400 Bad Request' in result, result
+        assert b'500' not in result, result
+
     def test_024_expect_100_continue(self):
         def wsgi_app(environ, start_response):
             if int(environ['CONTENT_LENGTH']) > 1024:
@@ -748,27 +755,29 @@ class TestHttpd(_TestBase):
         result = read_http(sock)
         self.assertEqual(result.status, 'HTTP/1.1 417 Expectation Failed')
         self.assertEqual(result.body, b'failure')
-        fd.write(
-            b'PUT / HTTP/1.1\r\nHost: localhost\r\nContent-length: 7\r\n'
-            b'Expect: 100-continue\r\n\r\ntesting')
-        fd.flush()
-        header_lines = []
-        while True:
-            line = fd.readline()
-            if line == b'\r\n':
-                break
-            else:
-                header_lines.append(line)
-        assert header_lines[0].startswith(b'HTTP/1.1 100 Continue')
-        header_lines = []
-        while True:
-            line = fd.readline()
-            if line == b'\r\n':
-                break
-            else:
-                header_lines.append(line)
-        assert header_lines[0].startswith(b'HTTP/1.1 200 OK')
-        assert fd.read(7) == b'testing'
+
+        for expect_value in ('100-continue', '100-Continue'):
+            fd.write(
+                'PUT / HTTP/1.1\r\nHost: localhost\r\nContent-length: 7\r\n'
+                'Expect: {}\r\n\r\ntesting'.format(expect_value).encode())
+            fd.flush()
+            header_lines = []
+            while True:
+                line = fd.readline()
+                if line == b'\r\n':
+                    break
+                else:
+                    header_lines.append(line)
+            assert header_lines[0].startswith(b'HTTP/1.1 100 Continue')
+            header_lines = []
+            while True:
+                line = fd.readline()
+                if line == b'\r\n':
+                    break
+                else:
+                    header_lines.append(line)
+            assert header_lines[0].startswith(b'HTTP/1.1 200 OK')
+            assert fd.read(7) == b'testing'
         fd.close()
         sock.close()
 
@@ -979,6 +988,67 @@ class TestHttpd(_TestBase):
         fd.close()
         sock.close()
 
+    def test_024d_expect_100_continue_with_eager_app_chunked(self):
+        def wsgi_app(environ, start_response):
+            # app knows it's going to do some time-intensive thing and doesn't
+            # want clients to time out, so it's protocol says to:
+            # * generally expect a successful status code,
+            # * be prepared to eat some whitespace that will get dribbled out
+            #   periodically, and
+            # * parse the final status from the response body.
+            environ['eventlet.minimum_write_chunk_size'] = 0
+            start_response('202 Accepted', [])
+
+            def resp_gen():
+                yield b' '
+                environ['wsgi.input'].read()
+                yield b' '
+                yield b'\n503 Service Unavailable\n\nOops!\n'
+
+            return resp_gen()
+
+        self.site.application = wsgi_app
+        sock = eventlet.connect(self.server_addr)
+        fd = sock.makefile('rwb')
+        fd.write(b'PUT /a HTTP/1.1\r\n'
+                 b'Host: localhost\r\nConnection: close\r\n'
+                 b'Transfer-Encoding: chunked\r\n'
+                 b'Expect: 100-continue\r\n\r\n')
+        fd.flush()
+
+        # Expect the optimistic response
+        header_lines = []
+        while True:
+            line = fd.readline()
+            if line == b'\r\n':
+                break
+            else:
+                header_lines.append(line.strip())
+        self.assertEqual(header_lines[0], b'HTTP/1.1 202 Accepted')
+
+        def chunkify(data):
+            return '{:x}'.format(len(data)).encode('ascii') + b'\r\n' + data + b'\r\n'
+
+        def expect_chunk(data):
+            expected = chunkify(data)
+            self.assertEqual(expected, fd.read(len(expected)))
+
+        # Can even see that initial whitespace
+        expect_chunk(b' ')
+
+        # Send message
+        fd.write(chunkify(b'some data'))
+        fd.write(chunkify(b''))  # end-of-message
+        fd.flush()
+
+        # Expect final response
+        expect_chunk(b' ')
+        expect_chunk(b'\n503 Service Unavailable\n\nOops!\n')
+        expect_chunk(b'')  # end-of-message
+
+        fd.close()
+        sock.close()
+
     def test_025_accept_errors(self):
         debug.hub_exceptions(True)
         listener = greensocket.socket()
@@ -1357,6 +1427,32 @@ class TestHttpd(_TestBase):
         self.assertEqual(read_content.wait(), b'ok')
         assert blew_up[0]
 
+    def test_aborted_post_io_error(self):
+        ran_post_req_hook = [False]
+
+        def post_req_hook(env):
+            ran_post_req_hook[0] = True
+
+        def early_responder(env, start_response):
+            env['eventlet.posthooks'] = [(post_req_hook, (), {})]
+            start_response('200 OK', [('Content-Type', 'text/plain')])
+            return ['ok']
+
+        self.site.application = early_responder
+        data = "\r\n".join(['PUT /somefile HTTP/1.1',
+                            'Transfer-Encoding: chunked',
+                            '',
+                            '20',
+                            'not 32 bytes'])
+        sock = eventlet.connect(self.server_addr)
+        sock.sendall(data.encode())
+        sock.close()
+        # Give the server a chance to wrap things up
+        eventlet.sleep(0.01)
+        # Unexpected EOF shouldn't kill the server;
+        # post-request processing should still happen
+        assert ran_post_req_hook[0]
+
     def test_exceptions_close_connection(self):
         def wsgi_app(environ, start_response):
             raise RuntimeError("intentional error")
@@ -1415,21 +1511,28 @@ class TestHttpd(_TestBase):
 
         self.site.application = wsgi_app
         sock = eventlet.connect(self.server_addr)
-        sock.sendall(b'GET /%E4%BD%A0%E5%A5%BD HTTP/1.1\r\nHost: localhost\r\nConnection: close\r\n\r\n')
+        # This is a properly-quoted request for the UTF-8 path /你好
+        sock.sendall(b'GET /%E4%BD%A0%E5%A5%BD HTTP/1.1\r\nHost: localhost\r\n\r\n')
         result = read_http(sock)
         assert result.status == 'HTTP/1.1 200 OK'
-        # that was only preparation, actual test below
+        # Like above, but the octets are reversed before being quoted,
+        # so the result should *not* be interpreted as UTF-8
+        sock.sendall(b'GET /%BD%A5%E5%A0%BD%E4 HTTP/1.1\r\nHost: localhost\r\nConnection: close\r\n\r\n')
+        result = read_http(sock)
+        assert result.status == 'HTTP/1.1 200 OK'
+
+        # that was only preparation, actual tests below
         # Per PEP-0333 https://www.python.org/dev/peps/pep-0333/#unicode-issues
         # in all WSGI environment strings application must observe either bytes in latin-1 (ISO-8859-1)
         # or unicode code points \u0000..\u00ff
-        # wsgi_decoding_dance from Werkzeug to emulate concerned application
         msg = 'Expected PATH_INFO to be a native string, not {0}'.format(type(g[0]))
         assert isinstance(g[0], str), msg
-        if six.PY2:
-            assert g[0] == u'/你好'.encode('utf-8')
-        else:
-            decoded = g[0].encode('latin1').decode('utf-8', 'replace')
-            assert decoded == u'/你好'
+        # Fortunately, WSGI strings have the same literal representation on both py2 and py3
+        assert g[0] == '/\xe4\xbd\xa0\xe5\xa5\xbd'
+
+        msg = 'Expected PATH_INFO to be a native string, not {0}'.format(type(g[1]))
+        assert isinstance(g[1], str), msg
+        assert g[1] == '/\xbd\xa5\xe5\xa0\xbd\xe4'
 
     @tests.skip_if_no_ipv6
     def test_ipv6(self):
@@ -1738,6 +1841,12 @@ class TestChunkedInput(_TestBase):
         elif pi == "/lines":
             for x in input:
                 response.append(x)
+        elif pi == "/readline":
+            response.extend(iter(input.readline, b''))
+            response.append(('\nread %d lines' % len(response)).encode())
+        elif pi == "/readlines":
+            response.extend(input.readlines())
+            response.append(('\nread %d lines' % len(response)).encode())
         elif pi == "/ping":
             input.read()
             response.append(b"pong")
@@ -1841,6 +1950,26 @@ class TestChunkedInput(_TestBase):
         self.assertEqual(read_http(fd).body, b'this is chunked\nline 2\nline3')
         fd.close()
 
+    def test_chunked_readline_from_input(self):
+        body = self.body()
+        req = "POST /readline HTTP/1.1\r\nContent-Length: %s\r\n" \
+              "transfer-encoding: Chunked\r\n\r\n%s" % (len(body), body)
+
+        fd = self.connect()
+        fd.sendall(req.encode())
+        self.assertEqual(read_http(fd).body, b'this is chunked\nline 2\nline3\nread 3 lines')
+        fd.close()
+
+    def test_chunked_readlines_from_input(self):
+        body = self.body()
+        req = "POST /readlines HTTP/1.1\r\nContent-Length: %s\r\n" \
+              "transfer-encoding: Chunked\r\n\r\n%s" % (len(body), body)
+
+        fd = self.connect()
+        fd.sendall(req.encode())
+        self.assertEqual(read_http(fd).body, b'this is chunked\nline 2\nline3\nread 3 lines')
+        fd.close()
+
     def test_chunked_readline_wsgi_override_minimum_chunk_size(self):
 
         fd = self.connect()
